@@ -29,7 +29,6 @@
 #include <gio/gio.h>
 #include <gtk/gtk.h>
 #include <webkit/webkit.h>
-#include <webkit/webkitwebresource.h>
 
 #include "yelp-debug.h"
 #include "yelp-docbook-document.h"
@@ -68,6 +67,10 @@ static void        popup_save_image               (GtkMenuItem        *item,
                                                    YelpView           *view);
 static void        popup_send_image               (GtkMenuItem        *item,
                                                    YelpView           *view);
+static void        popup_copy_code                (GtkMenuItem        *item,
+                                                   YelpView           *view);
+static void        popup_save_code                (GtkMenuItem        *item,
+                                                   YelpView           *view);
 static void        view_populate_popup            (YelpView           *view,
                                                    GtkMenu            *menu,
                                                    gpointer            data);
@@ -105,6 +108,10 @@ static void        document_callback              (YelpDocument       *document,
                                                    YelpDocumentSignal  signal,
                                                    YelpView           *view,
                                                    GError             *error);
+static gboolean    dom_node_is_name               (WebKitDOMNode      *node,
+                                                   gchar              *name);
+static gboolean    dom_node_has_class             (WebKitDOMNode      *node,
+                                                   gchar              *class);
 
 static const GtkActionEntry entries[] = {
     {"YelpViewPrint", GTK_STOCK_PRINT,
@@ -196,6 +203,9 @@ struct _YelpViewPrivate {
     gchar         *popup_link_uri;
     gchar         *popup_link_text;
     gchar         *popup_image_uri;
+    WebKitDOMNode *popup_code_node;
+    WebKitDOMNode *popup_code_title;
+    gchar         *popup_code_text;
 
     YelpViewState  state;
 
@@ -307,6 +317,7 @@ yelp_view_finalize (GObject *object)
     g_free (priv->popup_link_uri);
     g_free (priv->popup_link_text);
     g_free (priv->popup_image_uri);
+    g_free (priv->popup_code_text);
 
     g_free (priv->page_id);
     g_free (priv->root_title);
@@ -742,7 +753,7 @@ file_copied (GFile        *file,
 {
     GError *error = NULL;
     if (!g_file_copy_finish (file, res, &error)) {
-        GtkWidget *dialog = gtk_message_dialog_new (GTK_WIDGET_VISIBLE (data->window) ? data->window : NULL,
+        GtkWidget *dialog = gtk_message_dialog_new (gtk_widget_get_visible (GTK_WIDGET (data->window)) ? data->window : NULL,
                                                     GTK_DIALOG_DESTROY_WITH_PARENT,
                                                     GTK_MESSAGE_ERROR,
                                                     GTK_BUTTONS_OK,
@@ -828,6 +839,97 @@ popup_send_image (GtkMenuItem *item,
     priv->popup_image_uri = NULL;
 }
 
+static void
+popup_copy_code (GtkMenuItem *item,
+                 YelpView    *view)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    GtkClipboard *clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
+    gchar *content = webkit_dom_node_get_text_content (priv->popup_code_node);
+    gtk_clipboard_set_text (clipboard, content, -1);
+    g_free (content);
+}
+
+static void
+popup_save_code (GtkMenuItem *item,
+                 YelpView    *view)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    GtkWidget *dialog, *window;
+    gint res;
+
+    g_free (priv->popup_code_text);
+    priv->popup_code_text = webkit_dom_node_get_text_content (priv->popup_code_node);
+    if (!g_str_has_suffix (priv->popup_code_text, "\n")) {
+        gchar *tmp = g_strconcat (priv->popup_code_text, "\n", NULL);
+        g_free (priv->popup_code_text);
+        priv->popup_code_text = tmp;
+    }
+
+    for (window = gtk_widget_get_parent (GTK_WIDGET (view));
+         window && !GTK_IS_WINDOW (window);
+         window = gtk_widget_get_parent (window));
+
+    dialog = gtk_file_chooser_dialog_new (_("Save Code"),
+                                          GTK_WINDOW (window),
+                                          GTK_FILE_CHOOSER_ACTION_SAVE,
+                                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+                                          GTK_STOCK_SAVE, GTK_RESPONSE_OK,
+                                          NULL);
+    gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
+    if (priv->popup_code_title) {
+        gchar *filename = webkit_dom_node_get_text_content (priv->popup_code_title);
+        gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), filename);
+        g_free (filename);
+    }
+    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog),
+                                         g_get_user_special_dir (G_USER_DIRECTORY_DOCUMENTS));
+
+    res = gtk_dialog_run (GTK_DIALOG (dialog));
+
+    if (res == GTK_RESPONSE_OK) {
+        GError *error = NULL;
+        GFile *file = gtk_file_chooser_get_file (GTK_FILE_CHOOSER (dialog));
+        GFileOutputStream *stream = g_file_replace (file, NULL, FALSE,
+                                                    G_FILE_CREATE_NONE,
+                                                    NULL,
+                                                    &error);
+        if (stream == NULL) {
+            GtkWidget *dialog = gtk_message_dialog_new (gtk_widget_get_visible (window) ? GTK_WINDOW (window) : NULL,
+                                                        GTK_DIALOG_DESTROY_WITH_PARENT,
+                                                        GTK_MESSAGE_ERROR,
+                                                        GTK_BUTTONS_OK,
+                                                        "%s", error->message);
+            gtk_dialog_run (GTK_DIALOG (dialog));
+            gtk_widget_destroy (dialog);
+            g_error_free (error);
+        }
+        else {
+            /* FIXME: we should do this async */
+            GDataOutputStream *datastream = g_data_output_stream_new (G_OUTPUT_STREAM (stream));
+            if (!g_data_output_stream_put_string (datastream, priv->popup_code_text, NULL, &error)) {
+                GtkWidget *dialog = gtk_message_dialog_new (gtk_widget_get_visible (window) ? GTK_WINDOW (window) : NULL,
+                                                            GTK_DIALOG_DESTROY_WITH_PARENT,
+                                                            GTK_MESSAGE_ERROR,
+                                                            GTK_BUTTONS_OK,
+                                                            "%s", error->message);
+                gtk_dialog_run (GTK_DIALOG (dialog));
+                gtk_widget_destroy (dialog);
+                g_error_free (error);
+            }
+            g_object_unref (datastream);
+        }
+        g_object_unref (file);
+    }
+
+    priv->popup_code_node = NULL;
+    priv->popup_code_title = NULL;
+    g_free (priv->popup_code_text);
+    priv->popup_code_text = NULL;
+
+    gtk_widget_destroy (dialog);
+}
+
 static void
 view_populate_popup (YelpView *view,
                      GtkMenu  *menu,
@@ -839,6 +941,7 @@ view_populate_popup (YelpView *view,
     YelpViewPrivate *priv = GET_PRIV (view);
     GList *children;
     GtkWidget *item;
+    WebKitDOMNode *node, *cur, *link_node = NULL, *code_node = NULL, *code_title_node = NULL;
 
     children = gtk_container_get_children (GTK_CONTAINER (menu));
     while (children) {
@@ -851,7 +954,26 @@ view_populate_popup (YelpView *view,
     event = gtk_get_current_event ();
 
     result = webkit_web_view_get_hit_test_result (WEBKIT_WEB_VIEW (view), (GdkEventButton *) event);
-    g_object_get (result, "context", &context, NULL);
+    g_object_get (result,
+                  "context", &context,
+                  "inner-node", &node,
+                  NULL);
+    for (cur = node; cur != NULL; cur = webkit_dom_node_get_parent_node (cur)) {
+        if (dom_node_is_name (cur, "a"))
+            link_node = cur;
+
+        if (dom_node_is_name (cur, "div") && dom_node_has_class (cur, "code")) {
+            WebKitDOMNode *title;
+            code_node = cur;
+            title = webkit_dom_node_get_parent_node (cur);
+            if (title && dom_node_is_name (title, "div") && dom_node_has_class (title, "contents")) {
+                title = webkit_dom_node_get_previous_sibling (title);
+                if (title && dom_node_is_name (title, "div") && dom_node_has_class (title, "title")) {
+                    code_title_node = title;
+                }
+            }
+        }
+    }
 
     if (context & WEBKIT_HIT_TEST_RESULT_CONTEXT_LINK) {
         gchar *uri;
@@ -860,8 +982,47 @@ view_populate_popup (YelpView *view,
         priv->popup_link_uri = uri;
 
         g_free (priv->popup_link_text);
-        /* FIXME */
-        priv->popup_link_text = g_strdup (uri);
+        priv->popup_link_text = NULL;
+        /* FIXME: Handled space-separated class names, etc. See about a convenience API
+         * in WebKit, because this kind of processing in C really sucks.
+         */
+        if (link_node != NULL) {
+            WebKitDOMNode *child = webkit_dom_node_get_first_child (link_node);
+            gchar *tmp;
+            gint i, tmpi;
+            gboolean ws;
+            if (dom_node_is_name (child, "div") && dom_node_has_class (child, "linkdiv")) {
+                child = webkit_dom_node_get_first_child (child);
+                if (child && dom_node_is_name (child, "div") && dom_node_has_class (child, "title")) {
+                    priv->popup_link_text = webkit_dom_node_get_text_content (child);
+                }
+            }
+            if (priv->popup_link_text == NULL)
+                priv->popup_link_text = webkit_dom_node_get_text_content (link_node);
+
+            tmp = g_new0 (gchar, strlen(priv->popup_link_text) + 1);
+            ws = FALSE;
+            for (i = 0, tmpi = 0; priv->popup_link_text[i] != '\0'; i++) {
+                if (priv->popup_link_text[i] == ' ' || priv->popup_link_text[i] == '\n') {
+                    if (!ws) {
+                        tmp[tmpi] = ' ';
+                        tmpi++;
+                        ws = TRUE;
+                    }
+                }
+                else {
+                    tmp[tmpi] = priv->popup_link_text[i];
+                    tmpi++;
+                    ws = FALSE;
+                }
+            }
+            tmp[tmpi] = '\0';
+            g_free (priv->popup_link_text);
+            priv->popup_link_text = tmp;
+        }
+        else {
+            priv->popup_link_text = g_strdup (uri);
+        }
 
         if (g_str_has_prefix (priv->popup_link_uri, "mailto:")) {
             /* Not using a mnemonic because underscores are common in email
@@ -955,6 +1116,24 @@ view_populate_popup (YelpView *view,
         gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     }
 
+    if (code_node != NULL) {
+        item = gtk_separator_menu_item_new ();
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        priv->popup_code_node = code_node;
+        priv->popup_code_title = code_title_node;
+
+        item = gtk_menu_item_new_with_mnemonic (_("C_opy Code Block"));
+        g_signal_connect (item, "activate",
+                          G_CALLBACK (popup_copy_code), view);
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        item = gtk_menu_item_new_with_mnemonic (_("Save Code _Block As..."));
+        g_signal_connect (item, "activate",
+                          G_CALLBACK (popup_save_code), view);
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+    }
+
     g_object_unref (result);
     gdk_event_free (event);
     gtk_widget_show_all (GTK_WIDGET (menu));
@@ -1292,10 +1471,17 @@ uri_resolved (YelpUri  *uri,
         return;
     case YELP_URI_DOCUMENT_TYPE_NOT_FOUND:
         struri = yelp_uri_get_canonical_uri (uri);
-        error = g_error_new (YELP_ERROR, YELP_ERROR_NOT_FOUND,
-                             _("The URI ‘%s’ does point to a valid page."),
-                             struri);
-        g_free (struri);
+        if (struri != NULL) {
+            error = g_error_new (YELP_ERROR, YELP_ERROR_NOT_FOUND,
+                                 _("The URI ‘%s’ does not point to a valid page."),
+                                 struri);
+            g_free (struri);
+        }
+        else {
+            error = g_error_new (YELP_ERROR, YELP_ERROR_NOT_FOUND,
+                                 _("The URI does not point to a valid page."),
+                                 struri);
+        }
         view_show_error_page (view, error);
         return;
     case YELP_URI_DOCUMENT_TYPE_ERROR:
@@ -1574,3 +1760,37 @@ document_callback (YelpDocument       *document,
         view_show_error_page (view, error);
     }
 }
+
+static gboolean
+dom_node_is_name (WebKitDOMNode *node,
+                  gchar         *name)
+{
+    gboolean ret;
+    gchar *nodename = webkit_dom_node_get_node_name (node);
+    ret = g_str_equal (nodename, name);
+    g_free (nodename);
+    return ret;
+}
+
+static gboolean
+dom_node_has_class (WebKitDOMNode *node,
+                    gchar         *class)
+{
+    gboolean ret = FALSE;
+    WebKitDOMNamedNodeMap *map = webkit_dom_node_get_attributes (node);
+    WebKitDOMNode *attr = webkit_dom_named_node_map_get_named_item (map, "class");
+    if (attr) {
+        gchar *htmlclass = webkit_dom_node_get_text_content (attr);
+        gchar **classes = g_strsplit (htmlclass, " ", -1);
+        gint classi;
+        for (classi = 0; classes[classi] != NULL; classi++) {
+            if (g_str_equal (classes[classi], class)) {
+                ret = TRUE;
+                break;
+            }
+        }
+        g_strfreev (classes);
+        g_free (htmlclass);
+    }
+    return ret;
+}
